# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the proxy tunnel."""

from urlparse import urlparse

from twisted.internet import defer, protocol, reactor

from ubuntuone.devtools.testcases.squid import SquidTestCase

from tests.proxy import (
    FakeTransport,
    MockWebServer,
    SAMPLE_CONTENT,
    SIMPLERESOURCE,
)
from ubuntuone.proxy import tunnel_server
from ubuntuone.proxy.tunnel_server import CRLF


FAKE_SESSION_TEMPLATE = (
    "CONNECT {0.netloc} HTTP/1.0" + CRLF +
    "Header1: value1" + CRLF +
    "Header2: value2" + CRLF +
    CRLF +
    "GET {0.path} HTTP/1.0" + CRLF + CRLF
)


class DisconnectingProtocol(protocol.Protocol):
    """A protocol that just disconnects."""

    def connectionMade(self):
        """Upon connecting: just disconnect."""
        self.transport.loseConnection()


class DisconnectingClientFactory(protocol.ClientFactory):
    """A factory that fires a deferred on connection."""

    def __init__(self):
        """Initialize this instance."""
        self.connected = defer.Deferred()

    def buildProtocol(self, addr):
        """The connection was made."""
        proto = DisconnectingProtocol()
        if not self.connected.called:
            self.connected.callback(proto)
        return proto


class FakeProtocol(protocol.Protocol):
    """A protocol that forwards some data."""

    def __init__(self, factory, data):
        """Initialize this fake."""
        self.factory = factory
        self.data = data
        self.received_data = []

    def connectionMade(self):
        """Upon connection: send the stored data."""
        self.transport.write(self.data)

    def dataReceived(self, data):
        """Some data was received."""
        self.received_data.append(data)

    def connectionLost(self, reason):
        """The connection was lost, return the response."""
        response = "".join(self.received_data)
        if not self.factory.response.called:
            self.factory.response.callback(response)


class FakeClientFactory(protocol.ClientFactory):
    """A factory that forwards some data to the protocol."""

    def __init__(self, data):
        """Initialize this fake."""
        self.data = data
        self.response = defer.Deferred()

    def buildProtocol(self, addr):
        """The connection was made."""
        return FakeProtocol(self, self.data)


class TunnelIntegrationTestCase(SquidTestCase):
    """Basic tunnel integration tests."""

    timeout = 3

    @defer.inlineCallbacks
    def setUp(self):
        """Initialize this testcase."""
        yield super(TunnelIntegrationTestCase, self).setUp()
        self.ws = MockWebServer()
        self.addCleanup(self.ws.stop)
        self.dest_url = self.ws.get_iri().encode("utf-8") + SIMPLERESOURCE
        self.tunnel_server = tunnel_server.TunnelServer()
        self.addCleanup(self.tunnel_server.shutdown)

    def test_init(self):
        """The tunnel is started."""
        self.assertNotEqual(self.tunnel_server.port, 0)

    @defer.inlineCallbacks
    def test_accepts_connections(self):
        """The tunnel accepts incoming connections."""
        ncf = DisconnectingClientFactory()
        reactor.connectTCP("0.0.0.0", self.tunnel_server.port, ncf)
        yield ncf.connected

    @defer.inlineCallbacks
    def test_complete_connection(self):
        """Test from the tunnel server down."""
        url = urlparse(self.dest_url)
        client = FakeClientFactory(FAKE_SESSION_TEMPLATE.format(url))
        reactor.connectTCP("0.0.0.0", self.tunnel_server.port, client)
        response = yield client.response
        self.assertIn(SAMPLE_CONTENT, response)


class FakeClient(object):
    """A fake destination client."""

    protocol = None
    connection_result = defer.succeed(True)

    def connect(self, hostport):
        """Establish a connection with the other end."""
        return self.connection_result

    def write(self, data):
        """Write some data to the other end."""
        if data == 'GET /simpleresource HTTP/1.0\r\n\r\n':
            self.protocol.transport.write(SAMPLE_CONTENT)

    def stop(self):
        """Stop this fake client."""


class ServerTunnelProtocolTestCase(SquidTestCase):
    """Tests for the ServerTunnelProtocol."""

    @defer.inlineCallbacks
    def setUp(self):
        """Initialize this test instance."""
        yield super(ServerTunnelProtocolTestCase, self).setUp()
        self.ws = MockWebServer()
        self.addCleanup(self.ws.stop)
        self.dest_url = self.ws.get_iri().encode("utf-8") + SIMPLERESOURCE
        self.transport = FakeTransport()
        self.fake_client = FakeClient()
        self.proto = tunnel_server.ServerTunnelProtocol(
                                                lambda _: self.fake_client)
        self.fake_client.protocol = self.proto
        self.proto.transport = self.transport

    def test_broken_request(self):
        """Broken request."""
        self.proto.dataReceived("Broken request." + CRLF)
        self.assertTrue(self.transport.getvalue().startswith("HTTP/1.0 400 "),
                        "A broken request must fail.")

    def test_wrong_method(self):
        """Wrong method."""
        self.proto.dataReceived("GET http://slashdot.org HTTP/1.0" + CRLF)
        self.assertTrue(self.transport.getvalue().startswith("HTTP/1.0 405 "),
                        "Using a wrong method fails.")

    def test_invalid_http_version(self):
        """Invalid HTTP version."""
        self.proto.dataReceived("CONNECT 127.0.0.1:9999 HTTP/1.1" + CRLF)
        self.assertTrue(self.transport.getvalue().startswith("HTTP/1.0 505 "),
                        "Invalid http version is not allowed.")

    def test_connection_is_established(self):
        """The response code is sent."""
        expected = "HTTP/1.0 200 Proxy connection established" + CRLF
        self.proto.dataReceived("CONNECT 127.0.0.1:9999 HTTP/1.0" + CRLF * 2)
        self.assertTrue(self.transport.getvalue().startswith(expected),
                        "First line must be the response status")

    def test_connection_fails(self):
        """The connection to the other end fails, and it's handled."""
        error = tunnel_server.ConnectionError()
        self.patch(self.fake_client, "connection_result", defer.fail(error))
        expected = "HTTP/1.0 500 Connection error" + CRLF
        self.proto.dataReceived("CONNECT 127.0.0.1:9999 HTTP/1.0" + CRLF * 2)
        self.assertTrue(self.transport.getvalue().startswith(expected),
                        "The connection should fail at this point.")

    def test_headers_stored(self):
        """The request headers are stored."""
        expected = [
            ("Header1", "value1"),
            ("Header2", "value2"),
        ]
        self.proto.dataReceived("CONNECT 127.0.0.1:9999 HTTP/1.0" + CRLF +
                                "Header1: value1" + CRLF +
                                "Header2: value2" + CRLF + CRLF)
        self.assertEqual(self.proto.received_headers, expected)

    def test_successful_connect(self):
        """A successful connect thru the tunnel."""
        url = urlparse(self.dest_url)
        data = FAKE_SESSION_TEMPLATE.format(url)
        self.proto.dataReceived(data)
        lines = self.transport.getvalue().split(CRLF)
        self.assertEqual(lines[-1], SAMPLE_CONTENT)

    def test_header_split(self):
        """Test a header with many colons."""
        self.proto.header_line("key: host:port")
        self.assertIn("key", dict(self.proto.received_headers))


class FakeServerTunnelProtocol(object):
    """A fake ServerTunnelProtocol."""

    def __init__(self):
        """Initialize this fake tunnel."""
        self.response_received = defer.Deferred()

    def response_data_received(self, data):
        """Fire the response deferred."""
        if not self.response_received.called:
            self.response_received.callback(data)

    def remote_disconnected(self):
        """The remote server disconnected."""


class RemoteSocketTestCase(SquidTestCase):
    """Tests for the client that connects to the other side."""

    timeout = 3
    get_proxy_settings = lambda _: {}

    @defer.inlineCallbacks
    def setUp(self):
        """Initialize this testcase."""
        yield super(RemoteSocketTestCase, self).setUp()
        self.ws = MockWebServer()
        self.addCleanup(self.ws.stop)
        self.dest_url = self.ws.get_iri().encode("utf-8") + SIMPLERESOURCE

        self.addCleanup(tunnel_server.QNetworkProxy.setApplicationProxy,
                        tunnel_server.QNetworkProxy.applicationProxy())
        tunnel_server.QNetworkProxy.setApplicationProxy(
                        tunnel_server.build_proxy(self.get_proxy_settings()))

    def test_invalid_port(self):
        """A request with an invalid port fails with a 400."""
        protocol = tunnel_server.ServerTunnelProtocol(
                                                    tunnel_server.RemoteSocket)
        protocol.transport = FakeTransport()
        protocol.dataReceived("CONNECT 127.0.0.1:wrong_port HTTP/1.0" +
                              CRLF * 2)

        status_line = protocol.transport.getvalue()
        self.assertTrue(status_line.startswith("HTTP/1.0 400 "),
                        "The port must be an integer.")

    @defer.inlineCallbacks
    def test_connection_is_finished_when_stopping(self):
        """The client disconnects when requested."""
        fake_protocol = FakeServerTunnelProtocol()
        client = tunnel_server.RemoteSocket(fake_protocol)
        url = urlparse(self.dest_url)
        yield client.connect(url.netloc)
        yield client.stop()

    @defer.inlineCallbacks
    def test_stop_but_never_connected(self):
        """Stop but it was never connected."""
        fake_protocol = FakeServerTunnelProtocol()
        client = tunnel_server.RemoteSocket(fake_protocol)
        yield client.stop()

    @defer.inlineCallbacks
    def test_client_write(self):
        """Data written to the client is sent to the other side."""
        fake_protocol = FakeServerTunnelProtocol()
        client = tunnel_server.RemoteSocket(fake_protocol)
        self.addCleanup(client.stop)
        url = urlparse(self.dest_url)
        yield client.connect(url.netloc)
        client.write("GET /simpleresource HTTP/1.0" + CRLF * 2)
        yield self.ws.simple_resource.rendered

    @defer.inlineCallbacks
    def test_client_read(self):
        """Data received by the client is written into the transport."""
        fake_protocol = FakeServerTunnelProtocol()
        client = tunnel_server.RemoteSocket(fake_protocol)
        self.addCleanup(client.stop)
        url = urlparse(self.dest_url)
        yield client.connect(url.netloc)
        client.write("GET /simpleresource HTTP/1.0" + CRLF * 2)
        yield self.ws.simple_resource.rendered
        data = yield fake_protocol.response_received
        _headers, content = str(data).split(CRLF * 2, 1)
        self.assertEqual(content, SAMPLE_CONTENT)


class AnonProxyRemoteSocketTestCase(RemoteSocketTestCase):
    """Tests for the client going thru an anonymous proxy."""

    get_proxy_settings = RemoteSocketTestCase.get_nonauth_proxy_settings

    def parse_headers(self, raw_headers):
        """Parse the headers."""
        lines = raw_headers.split(CRLF)
        header_lines = lines[1:]
        headers_pairs = (l.split(":", 1) for l in header_lines)
        return dict((k.lower(), v.strip()) for k, v in headers_pairs)

    @defer.inlineCallbacks
    def test_verify_client_uses_proxy(self):
        """Verify that the client uses the proxy."""
        fake_protocol = FakeServerTunnelProtocol()
        client = tunnel_server.RemoteSocket(fake_protocol)
        self.addCleanup(client.stop)
        url = urlparse(self.dest_url)
        yield client.connect(url.netloc)
        client.write("GET /simpleresource HTTP/1.0" + CRLF * 2)
        yield self.ws.simple_resource.rendered
        data = yield fake_protocol.response_received
        raw_headers, _content = str(data).split(CRLF * 2, 1)
        self.parse_headers(raw_headers)


class AuthenticatedProxyRemoteSocketTestCase(AnonProxyRemoteSocketTestCase):
    """Tests for the client going thru an authenticated proxy."""

    get_proxy_settings = RemoteSocketTestCase.get_auth_proxy_settings
